################################################
#           MSc Course on Econometrics         #
#             Loukia Meligkotsidou             #
################################################

# linear regression under heteroscedasticity 
# time-series data
# y: nx1 vector of observations on the dependent variable
# X: nxk matrix of observations on k covariates

# Load data
y <- as.matrix(read.table("C:/Loukia/R_labs_Econometrics/Lab2/Y.txt"))
X <- as.matrix(read.table("C:/Loukia/R_labs_Econometrics/Lab2/X.txt"))
cor(X)

dimX <- dim(X)
n=dimX[1]
k=dimX[2]
X <- cbind(1, X)  # Add a column of ones for the intercept term

# Least squares estimation
XX <- t(X) %*% X
XXinv <- solve(XX)
Xy <- t(X) %*% y
bhat <- XXinv %*% Xy
e <- y - X %*% bhat
plot(1:n, e, main = "Residuals")

sigmahat <- sum(e^2) / (n - k - 1)
Vbhat <- sigmahat * XXinv
diag(Vbhat)  # Estimate of variance ignoring heteroscedasticity

# Goldfeld-Quandt test for heteroscedasticity
n0 <- 42  # Structural break at t=n0
y1 <- y[1:n0]
y2 <- y[(n0 + 1):n]
X1 <- X[1:n0, ]
X2 <- X[(n0 + 1):n, ]

bhat1 <- solve(t(X1) %*% X1) %*% t(X1) %*% y1  # LSE in the first segment
e1 <- y1 - X1 %*% bhat1
sigmahat1 <- sum(e1^2) / (n0 - k - 1)  # Estimated variance of the first segment

bhat2 <- solve(t(X2) %*% X2) %*% t(X2) %*% y2  # LSE in the second segment
e2 <- y2 - X2 %*% bhat2
sigmahat2 <- sum(e2^2) / (n - n0 - k - 1)  # Estimated variance of the second segment

GQ <- sigmahat2 / sigmahat1  # Test statistic
pvalueGQ <- 1 - pf(GQ, n - n0 - k - 1, n0 - k - 1)
pvalueGQ

# White test
Xauxiliary <- cbind(X, X[, 2:(k+1)]^2, apply(X[, 2:(k+1)], 1, prod))
yauxiliary <- e^2

ahat <- solve(t(Xauxiliary) %*% Xauxiliary) %*% t(Xauxiliary) %*% yauxiliary  # LSE for auxiliary regression
lm_model <- lm(yauxiliary ~ Xauxiliary)  # Fit auxiliary regression
R2 <- summary(lm_model)$r.squared  # Coefficient of determination
LM <- R2 * n  # Test statistic
pvalueW <- 1 - pchisq(LM, df = ncol(Xauxiliary) - 1)
pvalueW

# GLS estimator
z <- c(rep(0, n0), rep(1, n - n0))
Z <- cbind(1, z)
ahat <- solve(t(Z) %*% Z) %*% t(Z) %*% yauxiliary
variances<-as.vector(1 / (Z %*% ahat))
Omegainv <- matrix(diag(variances),ncol=n)
XOXinv <- solve(t(X) %*% Omegainv %*% X)
bhat_GLS <- XOXinv %*% t(X) %*% Omegainv %*% y

bhat
diag(Vbhat)  # Estimate of variance ignoring heteroscedasticity

bhat_GLS
diag(XOXinv)  # Estimate of variance taking account of heteroscedasticity